package com.jdblabs.timestamper.core

import com.jdbernard.util.SmartConfig
import groovyx.net.http.HTTPBuilder
import java.text.SimpleDateFormat

import static groovyx.net.http.ContentType.*
import static groovyx.net.http.Method.*

public class JDBLabsWebTimelineSource extends TimelineSource {

    private static final String CONFIG_UNAME = "web.username"
    private static final String CONFIG_PWD = "web.password"
    private static final String CONFIG_TIMELINE = "web.timelineId"

    private static isoDateFormat

    static {
        isoDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'")
        isoDateFormat.timeZone = TimeZone.getTimeZone("UTC")
    }

    private String username
    private String password
    private String timelineId

    private URI baseUri
    private HTTPBuilder http

    private Map entryHashes

    public JDBLabsWebTimelineSource(URI uri, SmartConfig config) {
        super(uri)

        // load web-timeline specific properties
        username = config.getProperty(CONFIG_UNAME, "")
        password = config.getProperty(CONFIG_PWD, "")
        timelineId = config.getProperty(CONFIG_TIMELINE, "")

        baseUri = new URI(uri.scheme + "://" + uri.authority + "/ts_api/")
        http = new HTTPBuilder(baseUri)

        // set some default error handlers
        http.handler.'500' = { resp, json ->
            throw new IOException("The web timeline reported an internal " +
                "error: ${json.error ?: 'no details available'}") }

        http.handler.failure = {resp, json ->
            throw new IOException("Unable to complete the operation: error " +
                "communicating to the web timeline.\n" +
                "${resp.statusLine}: ${json}") }

        // init our hash of known entries
        entryHashes = [:]
    }

    public Timeline read() {
        def timelineJSON
        def entryListJSON
        def startDate, endDate
        Timeline timeline

        // make sure we have a fresh session
        authenticate()

        // load the timeline information
        http.request(GET, JSON) {
            uri.path = "timelines/${username}/${timelineId}"

            response.'404' = { resp ->
                throw new IOException("No timeline '${timelineId}' for user " +
                    "'${username}'") }

            response.success = { resp, json -> timelineJSON = json }
        }

        timeline = new Timeline()

        // create start and end times
        startDate = Calendar.getInstance()
        startDate.timeInMillis = 0
        startDate = isoDateFormat.format(startDate.time)

        endDate = isoDateFormat.format(Calendar.getInstance().time)

        // load the timeline entries
        entryListJSON = http.get(
            path: "entries/${username}/${timelineId}",
            contentType: JSON,
            query: [
                byDate: true,
                startDate: startDate,
                endDate: endDate ] ) { resp, json -> json }

        entryListJSON.each { entry ->
            
            // parse and create the timeline marker
            def timestamp = isoDateFormat.parse(entry.timestamp)
            def marker = new TimelineMarker(timestamp, entry.mark, entry.notes)

            // add it to the timeline and our map of hashes
            timeline.addMarker(marker)
            entryHashes[fullHash(marker)] = entry.id
        }

        // return the created timeline
        return timeline
    }

    public void persist(Timeline t) {
        Map deletedEntries
        List newEntries
        
        // make sure we have a fresh session
        authenticate()

        // find differences since we last persisted
        deletedEntries = entryHashes.clone()    // shallow copy
        newEntries = []
        
        t.each { marker ->
            def hash = fullHash(marker)

            // this marker is present, remove from deletedEntries
            if (deletedEntries.containsKey(hash)) {
                deletedEntries.remove(hash) }

            // this marker is not present, add to newEntries
            else { newEntries << marker }
        }

        // delete all entries that used to be present but are not any longer
        deletedEntries.each { hash, entryId ->
            http.request(DELETE) {
                uri.path = "entries/${username}/${timelineId}/${entryId}"

                response.'404' = { resp -> entryHashes.remove(hash) }
                response.success = { resp -> entryHashes.remove(hash) }
            }
        }

        // add all new entries
        newEntries.each { entry ->

            def entryBody = [
                mark: entry.mark,
                notes: entry.notes,
                timestamp: isoDateFormat.format(entry.timestamp)
            ]

            http.request(POST, JSON) {
                uri.path = "entries/${username}/${timelineId}"
                requestContentType = JSON
                body = entryBody

                response.success = { entryHashes.put(fullHash(entry), entry.id) }
            }
        }
    }

    public boolean isAuthenticated() { return false; }

    public void authenticate() throws AuthenticationException {
        http.request(POST, JSON) { req ->
            uri.path = '/ts_api/login'
            requestContentType = JSON
            body = [ username: this.username,
                    password: this.password ]

            response.'401' = { resp ->
                throw new AuthenticationException(
                    "Unable to connect to ${baseUri}: " +
                    "Invalid username/password combination.") }
        }
    }

    protected int fullHash(TimelineMarker tm) {
        return 61 * tm.hashCode() + (tm.mark != null ? tm.mark.hashCode() : 0);
    }
}
